#include <graph/graph.h>
#include <graph/maxmatch.h>
#include <graph/dc.h>
#include <util/input_generator.h>
#include <util/stopwatch.h>
#include <util/assert.h>
#include <memory/cache.h>
#include <typeinfo>

using namespace Memory;
using namespace Allocator;
using namespace std;

namespace {

typedef Disk<DiskCache<8192, 4096> > DCache;

class Mark : public Graph::Vertex {
	bool enabled;
public:
	Mark() :
		enabled(false)
	{ }
	void setMark(bool enabled) { this->enabled = enabled; }
	bool isMarkSet() const { return enabled; }
};

template<class Graph, class Pool, class Function> void singleTest(Graph &graph, const char *name, size_t size, Pool &allocator, Function function) {
	Util::Stopwatch watch(name, graph.name, Graph::StorageType::name(), "void", size);
	typename Pool::State state(allocator);
	watch.start();
	Pointer<DCache, ::Graph::Edge> result(function(graph, allocator));
	watch.stop();
	allocator.restore(state);
}

typedef Graph::Graph<DCache, Mark> G;

void test(Pool<DCache> &allocator, size_t size, G::Generator generator, int mult = 0) {
	Pool<DCache>::State state(allocator);
	DCache::resize(524288); // Allow more memory for faster graph generation
	G graph(G::generate(allocator, generator, size, mult));
	DCache::restoreSize(); // But the tests should not have it
	singleTest(graph, "maxMatchClassic", size, allocator, Graph::maxMatchClassic<G>);
	if (size < 956432250) {
		singleTest(graph, "maxMatchBRT", size, allocator, Graph::maxMatchBRT<G>);
	}
	singleTest(graph, "maxMatchDC", size, allocator, Graph::maxMatchDC<G>);
	allocator.restore(state);
}

}

int main() {
	Pool<DCache> allocator;
	size_t size(TEST_SIZE);
	for (double s(1); s < size; s *= 1.5) {
		test(allocator, s, G::Count, 3);
	}
	return 0;
}
