#include "sort.h"
#include <memory/cache.h>

#include <sort/qsort.h>
#include <sort/funnel.h>
#include <sort/heap.h>
#include <sort/merge.h>

namespace {

template<class Storage, class T> void runSorts(Pointer<Storage, T> input, Pointer<Storage, T> expected, Pointer<Storage, T> buffer, size_t size, const char *name, bool runThreaded) {
	RUN(qsort);
	RUN(funnel);
	RUN(funnelHybrid);
	RUN(funnelFullHybrid);
	RUN(heap);
	RUN(mergeStack);
	RUN(mergeFifo);
	if (runThreaded) {
		RUN(qsortThreaded);
		RUN(funnelThreaded);
		RUN(funnelHybridThreaded);
		RUN(funnelFullHybridThreaded);
		RUN(mergeStackThreaded);
		RUN(mergeFifoThreaded);
	}
}

}

int main() {
	vector<size_t> sizes;
	for (double size(TEST_SIZE); size >= 1.0; size /= 1.5)
		sizes.push_back(size);
	for (vector<size_t>::const_reverse_iterator s(sizes.rbegin()); s != sizes.rend(); s ++) {
		doTests<Core, int>(*s);
		doTests<Core, char>(*s);
		doTests<Core, double>(*s);
		doTests<Core, BigItem<16> >(*s);
		doTests<Core, BigItem<56> >(*s);
		doTests<Core, BigItem<128> >(*s);
		doTests<Core, BigItem<129> >(*s);
		doTests<Core, BigItem<200, double> >(*s);
	}
	return 0;
}
