#include <memory/memory.h>
#include <memory/mapping.h>
#include <memory/pointer.h>
#include <util/input_generator.h>
#include <util/bigitem.h>
#include <util/stopwatch.h>
#include <util/assert.h>
#include <cstdlib>

#include <sort/common.h>

using namespace Memory;
using namespace Util::Generator;
using namespace Util;
using namespace std;
using namespace Algorithm::Sort;

namespace {

template<class Storage, class T> void runSort(Pointer<Storage, T> input, Pointer<Storage, T> expected, Pointer<Storage, T> buffer, size_t size, const char *name, void (*sorter)(Pointer<Storage, T>, size_t), const char *algoname) {
#ifdef HUGE_INPUT
	generate(input, random<T>, size);
#else
	if (buffer != input)
		for (size_t i(0); i < size; ++ i)
			buffer[i] = input[i];
#endif
	Stopwatch watch(algoname, name, Storage::name(), typeid(T).name(), size);
	watch.start();
	sorter(buffer, size);
	watch.stop();
#ifndef NOCHECK
	for (size_t i(0); i < size; ++ i)
		ASSERT(buffer[i] == expected[i]);
#endif
}

template<class Storage, class T> void runSorts(Pointer<Storage, T> input, Pointer<Storage, T> expected, Pointer<Storage, T> buffer, size_t size, const char *name, bool runThreaded);

template<class Base> int compare(const void *b1, const void *b2) {
	const Base *_1(static_cast<const Base *>(b1)), *_2(static_cast<const Base *>(b2));
	if (*_1 < *_2)
		return -1;
	else if (*_1 > *_2)
		return 1;
	else
		return 0;
}

template<class Storage, class T> void doTests(size_t size, bool runThreaded = true, bool sizeLimit = true) {
#ifdef HUGE_INPUT
	const size_t size_mult(1), size_inc(0);
#else
	const size_t size_mult(3), size_inc(size);
#endif
	if (sizeLimit && size * sizeof(T) > TEST_SIZE)
		// Test too big to fit into size/memory
		return;
	Mapping<Storage> mapping(size_mult * size * sizeof(T), "Test input");
	Pointer<Storage, T> input(mapping), expected(input + size_inc), buffer(input + 2 * size_inc);
#ifndef HUGE_INPUT
#ifndef RANDOM_ONLY
	memset(input, 0, (size + size_inc) * sizeof(T));
	runSorts(input, expected, buffer, size, "zeroes", runThreaded);
#endif
	printf("OTHER:\tGenerating input of size %zd in %s of type %s ...\n", size, Storage::name(), typeid(T).name());
	fflush(stdout);
	for (size_t i(0); i < size; ++ i)
		random(input[i]);
#ifndef NOCHECK
	memmove(expected, input, size * sizeof(T));
	qsort(expected, size, sizeof(T), compare<T>);
#endif
#endif
	runSorts(input, expected, buffer, size, "random", runThreaded);
#ifndef HUGE_INPUT
#ifndef RANDOM_ONLY
	runSorts(expected, expected, buffer, size, "sorted", runThreaded);
#endif
#endif
}

}

#define RUN(algo) runSort(input, expected, buffer, size, name, algo, #algo)
