#ifndef GRAPH_CLASSIC_H
#define GRAPH_CLASSIC_H

#include "graph.h"
#include <util/stack.h>
#include <util/queue.h>

namespace Graph {
namespace Classic {

namespace Internal {

template<class Storage> struct Global {
	static Util::Stack<Storage, size_t> stack;
	static Util::Queue<Storage, size_t> *queue;
	static size_t qsize;
	static bool used;
};

template<class Storage> Util::Stack<Storage, size_t> Global<Storage>::stack;
template<class Storage> Util::Queue<Storage, size_t> *Global<Storage>::queue(NULL);
template<class Storage> size_t Global<Storage>::qsize(0);
template<class Storage> bool Global<Storage>::used(false);

template<class Graph, class Container> Memory::Pointer<typename Graph::StorageType, size_t> walk(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, Container &container, bool reverse, size_t origin) {
	Memory::Pointer<typename Graph::StorageType, size_t> result(allocator[graph.n]);
	container.push(origin);
	size_t pos(0);
	while (!container.empty()) {
		size_t vindex(container.pop());
		if (graph.vertices[vindex].isMarkSet())
			continue;
		graph.vertices[vindex].setMark(true);
		result[pos ++] = vindex;
		if (reverse) {
			for (size_t i(graph.vertices[vindex + 1].firstEdge - 1); i >= graph.vertices[vindex].firstEdge && i < graph.m * 2; -- i)
				// Check even for underflow of size_t
				container.push(graph.edges[i].vertices[1]);
		} else {
			for (size_t i(graph.vertices[vindex].firstEdge); i < graph.vertices[vindex + 1].firstEdge; ++ i)
				container.push(graph.edges[i].vertices[1]);
		}
	}
	result[pos] = graph.n + 1;
	for (size_t i(0); i < pos; ++ i)
		graph.vertices[result[i]].setMark(false);
	return result;
}

}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> dfs(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	typedef typename Graph::StorageType Storage;
	using namespace Internal;
	ASSERT(!Global<Storage>::used);
	Global<Storage>::used = true;
	Memory::Pointer<Storage, size_t> result(walk(graph, allocator, Global<Storage>::stack, true, origin));
	Global<Storage>::used = false;
	ASSERT(Global<Storage>::stack.empty());
	return result;
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> bfs(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	Util::Queue<typename Graph::StorageType, size_t> queue(graph.m * 2 + 1);
	return Internal::walk(graph, allocator, queue, false, origin);
	/*
	typedef typename Graph::StorageType Storage;
	using namespace Internal;
	ASSERT(!Global<Storage>::used);
	Global<Storage>::used = true;
	if (graph.m * 2 + 1 > Global<Storage>::qsize) {
		delete Global<Storage>::queue;
		Global<Storage>::queue = new Util::Queue<Storage, size_t>(graph.m * 2 + 1);
		Global<Storage>::qsize = graph.m * 2 + 1;
	} else {
		Global<Storage>::queue->reset();
	}
	Memory::Pointer<Storage, size_t> result(walk(graph, allocator, *Global<Storage>::queue, true, origin));
	Global<Storage>::used = false;
	return result;
	*/
}

}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> dfsClassic(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	return Classic::dfs(graph, allocator, origin);
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> bfsClassic(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	return Classic::bfs(graph, allocator, origin);
}

}

#endif
