#ifndef GRAPH_COMPONENTS_H
#define GRAPH_COMPONENTS_H

#include "graph.h"
#include "classic.h"
#include "treeish.h"
#include "sort-bfs.h"
#include <sort/sort.h>

namespace Graph {

template<class Graph, class Search, class Stack = Util::Stack<typename Graph::StorageType, Algorithm::Sort::QSort::Range> > Memory::Pointer<typename Graph::StorageType, size_t> componentsClassic(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, Search search, Stack *stack = NULL) {
	typedef typename Graph::StorageType Storage;
	typedef Memory::Pointer<Storage, size_t> Sizes;
	Sizes result(allocator[graph.n]);
	typedef typename Memory::Allocator::Pool<Storage>::State State;
	State original(allocator);
	for (size_t i(0); i < graph.n; ++ i)
		result[i] = graph.n + 1;
	size_t nextEmpty(0);
	size_t nextComponent(0);
	bool hadStack(stack);
	if (!hadStack)
		stack = new Stack;
	while (nextEmpty < graph.n) {
		Sizes scan(search(graph, allocator, nextEmpty));
		size_t end(0);
		while (end < graph.n && scan[end] != graph.n + 1)
			++ end;
		Algorithm::Sort::sort(scan, end, stack);
		for (size_t i(0); i < end; ++ i)
			result[scan[i]] = nextComponent;
		++ nextComponent;
		allocator.restore(original);
		while (nextEmpty < graph.n && result[nextEmpty] != graph.n + 1)
			++ nextEmpty;
	}
	if (!hadStack)
		delete stack;
	return result;
}

template<class Graph, class Container> Memory::Pointer<typename Graph::StorageType, size_t> componentsBRT(Graph &graph, Container &container, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	typedef typename Graph::StorageType Storage;
	typedef Memory::Pointer<Storage, size_t> Sizes;
	Sizes result(allocator[graph.n]);
	typedef typename Memory::Allocator::Pool<Storage>::State State;
	State original(allocator);
	Treeish::Internal::WalkState<Graph> state(graph, allocator);
	Sizes scan(allocator[graph.n]);
	for (size_t i(0); i < graph.n; ++ i)
		result[i] = graph.n + 1;
	size_t nextEmpty(0);
	size_t nextComponent(0);
	Util::Stack<typename Graph::StorageType, Algorithm::Sort::QSort::Range> stack;
	while (nextEmpty < graph.n) {
		Treeish::Internal::walkInternal(graph, container, state, scan, nextEmpty);
		size_t end(0);
		while (end < graph.n && scan[end] != graph.n + 1)
			++ end;
		Algorithm::Sort::sort(scan, end, &stack);
		for (size_t i(0); i < end; ++ i)
			result[scan[i]] = nextComponent;
		++ nextComponent;
		while (nextEmpty < graph.n && result[nextEmpty] != graph.n + 1)
			++ nextEmpty;
	}
	allocator.restore(original);
	return result;
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> componentsDfs(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	Util::Stack<typename Graph::StorageType, size_t> stack;
	return componentsClassic(graph, allocator, Classic::dfs<Graph>, &stack);
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> componentsBfs(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	Util::Stack<typename Graph::StorageType, size_t> stack;
	return componentsClassic(graph, allocator, Classic::bfs<Graph>, &stack);
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> componentsDfsBRT(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	Util::Stack<typename Graph::StorageType, size_t> stack;
	return componentsBRT(graph, stack, allocator);
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> componentsBfsBRT(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	Util::Queue<typename Graph::StorageType, size_t> queue(graph.n);
	return componentsBRT(graph, queue, allocator);
}

template<class Graph> class BfsSort {
private:
	typedef Util::Stack<typename Graph::StorageType, Algorithm::Sort::QSort::Range> Stack;
	Stack &stack;
public:
	Memory::Pointer<typename Graph::StorageType, size_t> operator ()(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) const {
		return Trick::sortBfs(graph, allocator, origin, &stack);
	}
	BfsSort(Stack &stack_) :
		stack(stack_)
	{ }
};

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> componentsBfsSort(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	typedef Util::Stack<typename Graph::StorageType, Algorithm::Sort::QSort::Range> Stack;
	Stack stack;
	BfsSort<Graph> sort(stack);
	return componentsClassic(graph, allocator, sort, &stack);
}

}

#endif
