#ifndef GRAPH_SORT_BFS_H
#define GRAPH_SORT_BFS_H

#include "graph.h"
#include <util/stack.h>
#include <util/assert.h>

namespace Graph {
namespace Trick {

template<class Graph, class Stack> Memory::Pointer<typename Graph::StorageType, size_t> sortBfs(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0, Stack *stack = NULL) {
	ASSERT(origin < graph.n);
	Memory::Pointer<typename Graph::StorageType, size_t> result(allocator[graph.n]);
	typename Memory::Allocator::Pool<typename Graph::StorageType>::State state(allocator);
	size_t pos(0);
	result[pos ++] = origin;
	size_t previous = 0;
	size_t current = 0;
	while (pos != current) {
		size_t expected((graph.n - 1) * (pos - current));
		if (expected > graph.m * 2)
			expected = graph.m * 2;
		Memory::Pointer<typename Graph::StorageType, size_t> layer(allocator[expected]);
		// Generate all neighboring vertices
		size_t npos(0);
		for (size_t i(current); i < pos; ++ i)
			for (size_t j(graph.vertices[result[i]].firstEdge); j < graph.vertices[result[i] + 1].firstEdge; ++ j) {
				ASSERT(npos < expected);
				layer[npos ++] = graph.edges[j].vertices[1];
			}
		// Sort them
		Algorithm::Sort::sort(layer, npos, stack);
		// Take the unique ones that are not in any previous layer
		size_t indices[3] = { previous, current, pos - 1 };
		for (size_t i(0); i < npos; ++ i) {
			// Move the indices so we are at high enough values
			for (size_t j(0); j < 2; ++ j)
				while (result[indices[j]] < layer[i] && indices[j] < indices[j + 1])
					indices[j] ++;
			// Did we see it in any of the three current layers already?
			for (size_t j(0); j < 3; ++ j)
				if (result[indices[j]] == layer[i])
					goto DUP;
			result[++ indices[2]] = layer[i];
			DUP:;
		}
		// Set the new ranges
		previous = current;
		current = pos;
		pos = indices[2] + 1;
		// And return memory so we can just take it again
		allocator.restore(state);
	}
	result[pos] = graph.n + 1;
	return result;
}

}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> bfsSort(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	Util::Stack<typename Graph::StorageType, Algorithm::Sort::QSort::Range> stack;
	return Trick::sortBfs(graph, allocator, origin, &stack);
}

}

#endif
