#ifndef POINTER_H
#define POINTER_H

#include "memory.h"

#include <cstdlib>

namespace Memory {

template<class Backend> class Mapping;
template<class Backend, class Target> class Pointer;

namespace Impl {

template<class Backend> class PointerBase;

template<class Cache> class PointerBase<Disk<Cache> > {
protected:
	typedef typename Cache::Id CacheId;
	CacheId id;
	PointerBase(const CacheId &id_ = CacheId()) : id(id_) {}
	template<class Target> void touch(Target *p, size_t size) { Cache::touch(p, size, id); }
public:
	CacheId getCacheId() const { return id; }
};

}

template<class Backend, class Target> class Pointer : public Impl::PointerBase<Backend> {
public:
	Pointer() : data(NULL) { }
	typedef Pointer<Backend, Target> Self;
	template<class T> static Self casted(Pointer<Backend, T> &other) {
		return Self(static_cast<Target *>(static_cast<void *>(other.get())), other.getCacheId());
	}
	Target &operator *() { touch(); return *data; }
	Target *operator ->() { touch(); return data; }
	operator void *() { return data; }
	Target *get() { touch(); return data; }
	Self operator +(size_t offset) { return Self(data + offset, Impl::PointerBase<Backend>::getCacheId()); }
	Self operator -(size_t offset) { return Self(data + offset, Impl::PointerBase<Backend>::getCacheId()); }
	const Self &operator +=(size_t offset) { data += offset; return *this; }
	const Self &operator -=(size_t offset) { data -= offset; return *this; }
	Self &operator ++() { ++ data; return *this; }
	Self &operator --() { ++ data; return *this; }
	Self operator ++(int) { ++ (*this); return (*this) - 1; }
	Self operator --(int) { -- (*this); return (*this) + 1; }
	Target &operator [](size_t index) { return *((*this) + index); }
	bool operator ==(const Self &other) const { return data == other.data; }
	bool operator !=(const Self &other) const { return data != other.data; }
	bool operator <(const Self &other) const { return data < other.data; }
	bool operator >(const Self &other) const { return data > other.data; }
	bool operator <=(const Self &other) const { return data <= other.data; }
	bool operator >=(const Self &other) const { return data >= other.data; }
	bool null() const { return data == NULL; }
private:
	friend class Mapping<Backend>;
	typedef typename Impl::PointerBase<Backend>::CacheId CacheId;
	Pointer(Target *raw, CacheId id = CacheId()) : Impl::PointerBase<Backend>(id), data(raw) { }
	Target *data;
	void touch() {
		Impl::PointerBase<Backend>::touch(data, sizeof(Target));
	}
};

template<class Target> class Pointer<Core, Target> {
public:
	Pointer() : data(NULL) { }
	typedef Pointer<Core, Target> Self;
	template<class T> static Self casted(Pointer<Core, T> &other) {
		return Self(static_cast<Target *>(static_cast<void *>(other.get())));
	}
	Target &operator *() { return *data; }
	Target *operator ->() { return data; }
	operator void *() { return data; }
	Target *get() { return data; }
	Self operator +(size_t offset) { return Self(data + offset); }
	Self operator -(size_t offset) { return Self(data + offset); }
	const Self &operator +=(size_t offset) { data += offset; return *this; }
	const Self &operator -=(size_t offset) { data -= offset; return *this; }
	Self &operator ++() { ++ data; return *this; }
	Self &operator --() { ++ data; return *this; }
	Self operator ++(int) { ++ (*this); return (*this) - 1; }
	Self operator --(int) { -- (*this); return (*this) + 1; }
	Target &operator [](size_t index) { return *((*this) + index); }
	bool operator ==(const Self &other) const { return data == other.data; }
	bool operator !=(const Self &other) const { return data != other.data; }
	bool operator <(const Self &other) const { return data < other.data; }
	bool operator >(const Self &other) const { return data > other.data; }
	bool operator <=(const Self &other) const { return data <= other.data; }
	bool operator >=(const Self &other) const { return data >= other.data; }
	bool null() const { return data == NULL; }
private:
	friend class Mapping<Core>;
	Pointer(Target *raw) : data(raw) { }
	Target *data;
};

}

#endif
