#ifndef SORT_QSORT_H
#define SORT_QSORT_H

#include "common.h"

#include <memory/pointer.h>
#include <util/stack.h>
#include <thread/thread.h>

namespace Algorithm {
namespace Sort {

namespace QSort {

struct Range {
	Range(size_t left_, size_t right_) :
		left(left_),
		right(right_)
	{ }
	size_t left, right;
	size_t length() const {
		if (left >= right)
			return 0;
		else
			return right - left;
	}
};

template<class Storage, class Type, class Stack> void qsort(Memory::Pointer<Storage, Type> array, size_t length, size_t threadCount, Stack &stack);

template<class Storage, class Type> class SubTask {
	private:
		Memory::Pointer<Storage, Type> array;
		size_t size;
		size_t threadCount;
	public:
		SubTask(Memory::Pointer<Storage, Type> array_, size_t size_, size_t threadCount_) :
			array(array_),
			size(size_),
			threadCount(threadCount_)
		{ }
		void operator ()() {
			Util::Stack<Storage, Range> stack;
			qsort<Storage, Type, Util::Stack<Storage, Range> >(array, size, threadCount, stack);
		}
};

template<class Storage, class Type, class Stack> void qsort(Memory::Pointer<Storage, Type> array, size_t length, size_t threadCount, Stack &stack) {
	using namespace QSort;
	stack.push(Range(0, length));
	while (!stack.empty()) {
		Range range(stack.pop());
		if (range.length() <= 10) {
			Common::bubble(array + range.left, range.length());
			continue;
		}
		size_t left(range.left), right(range.right - 1), middle((left + right) / 2);
		if (array[middle] < array[left]) {
			Type tmp(array[left]);
			array[left] = array[middle];
			array[middle] = tmp;
		}
		if (array[right] < array[middle]) {
			Type tmp(array[right]);
			array[right] = array[middle];
			array[middle] = tmp;
			if (array[middle] < array[left]) {
				Type tmp(array[left]);
				array[left] = array[middle];
				array[middle] = tmp;
			}
		}
		Type pivot(array[middle]);
		do {
			while (array[left] < pivot)
				left ++;
			while (array[right] > pivot)
				right --;
			if (left < right) {
				Type tmp(array[left]);
				array[left] = array[right];
				array[right] = tmp;
				left ++;
				right --;
			}
		} while (left < right);
		if (left > right) {
			right ++;
			left --;
		} else {
			if (pivot > array[left])
				right ++;
			else
				left --;
		}
		if (threadCount > 1 && length > 65536) {
			size_t first(Thread::enqueue(SubTask<Storage, Type>(array, left + 1, threadCount / 2)));
			size_t second(Thread::enqueue(SubTask<Storage, Type>(array + right, length - right, threadCount -  threadCount / 2)));
			Thread::wait(first);
			Thread::wait(second);
		} else {
			Range first(range.left, left + 1);
			Range second(right, range.right);
			if (first.length() > second.length()) {
				stack.push(first);
				stack.push(second);
			} else {
				stack.push(second);
				stack.push(first);
			}
		}
	}
}

}

template<class Storage, class Type> void qsort(Memory::Pointer<Storage, Type> array, size_t length) {
	Util::Stack<Storage, QSort::Range> stack;
	QSort::qsort<Storage, Type>(array, length, 1, stack);
}

template<class Storage, class Type> void qsortThreaded(Memory::Pointer<Storage, Type> array, size_t length) {
	Util::Stack<Storage, QSort::Range> stack;
	QSort::qsort<Storage, Type>(array, length, THREAD_COUNT * THREAD_COUNT, stack);
}

}
}

#endif
