#ifndef MUTEX_LOCK_H
#define MUTEX_LOCK_H

#include <pthread.h>
#include <util/assert.h>

namespace Thread {

class Lock {
private:
	Lock(const Lock &other);
	Lock &operator =(const Lock &);
	pthread_mutex_t mutex;
public:
	Lock() {
		int result(pthread_mutex_init(&mutex, NULL));
		ASSERT(result == 0);
	}
	~ Lock() {
		int result(pthread_mutex_destroy(&mutex));
		// FIXME: Do proper shutdow, this now dies at exit, because it's locked
		//ASSERT(result == 0);
	}
	void lock() {
		int result(pthread_mutex_lock(&mutex));
		ASSERT(result == 0);
	}
	void unlock() {
		int result(pthread_mutex_unlock(&mutex));
		ASSERT(result == 0);
	}
	pthread_mutex_t *osMutex() { return &mutex; }
};

class Scope {
private:
	Lock &lock;
	Scope(const Scope &other);
	Scope &operator =(const Scope &other);
public:
	Scope(Lock &lock_) :
		lock(lock_)
	{
		lock.lock();
	}
	~ Scope() {
		lock.unlock();
	}
};

}

#endif
