#ifndef THREAD_THREAD_H
#define THREAD_THREAD_H

#include <cstdlib>
#ifdef DEBUG_THREADS
#include <cstdio>
#include <typeinfo>
#endif

namespace Thread {

namespace Impl {

class BaseTask {
	public:
		size_t tid;
		virtual ~BaseTask() { }
		virtual void run() = 0;
};

template<class Functor> class Task : public BaseTask {
	private:
		Functor functor;
	public:
		Task(Functor functor_) : functor(functor_) { }
		virtual void run() {
#ifdef DEBUG_THREADS
			fprintf(DEBUG_THREADS, "DEBUG: Running task %s (%zd)\n", typeid(Functor).name(), tid);
#endif
			functor();
#ifdef DEBUG_THREADS
			fprintf(DEBUG_THREADS, "DEBUG: Task %s (%zd) finished\n", typeid(Functor).name(), tid);
#endif
		}
};

size_t enqueue(BaseTask *task);

}

template<class Functor> size_t enqueue(Functor functor) {
	return Impl::enqueue(new Impl::Task<Functor>(functor));
}

// You can wait for each task only once, either in wait or waitAll
void wait(size_t tid);
void waitAll();
// Number of tasks running
size_t running();
// Number of tasks waiting for a slot
size_t waiting();
// Number of unused slots (if >0, nothing is waiting and new tasks can be started)
size_t unused();

}

#endif
