#ifndef UTIL_COUNTITEM_H
#define UTIL_COUNTITEM_H

#include "input_generator.h"

namespace Util {

template<class Base> class CountItem {
	private:
		Base data;
		static size_t cmpcount;
		static size_t assigncount;
		static size_t mulcount;
		static size_t sumcount;
	public:
		CountItem() {}
		CountItem(const CountItem<Base>& other) :
			data(other.data)
		{
			++ assigncount;
		}
		CountItem(const Base& other) :
			data(other)
		{
			++ assigncount;
		}
		CountItem &operator =(const CountItem<Base>& other) {
			++ assigncount;
			data = other.data;
			return *this;
		}
		bool operator <(const CountItem<Base> &other) const { ++ cmpcount; return data < other.data; }
		bool operator >(const CountItem<Base> &other) const { ++ cmpcount; return data > other.data; }
		bool operator ==(const CountItem<Base> &other) const { ++ cmpcount; return data == other.data; }
		bool operator >=(const CountItem<Base> &other) const { ++ cmpcount; return data >= other.data; }
		bool operator <=(const CountItem<Base> &other) const { ++ cmpcount; return data <= other.data; }
		CountItem<Base> operator *(const CountItem<Base>& other) const {
			++ mulcount;
			return CountItem<Base>(data * other.data);
		}
		CountItem<Base> operator +(const CountItem<Base>& other) const {
			++ sumcount;
			return CountItem<Base>(data + other.data);
		}
		CountItem<Base>& operator +=(const CountItem<Base>& other) {
			++ sumcount;
			++ assigncount;
			return *this;
		}
		static void reset() { cmpcount = 0; assigncount = 0; mulcount = 0; sumcount = 0; }
		void random() {
			Generator::random(data);
		};
		static size_t getCmpCount() { return cmpcount; }
		static size_t getAssignCount() { return assigncount; }
		static size_t getMultiplyCount() { return mulcount; }
		static size_t getSumCount() { return sumcount; }
};

template<class Base> size_t CountItem<Base>::cmpcount(0);
template<class Base> size_t CountItem<Base>::assigncount(0);
template<class Base> size_t CountItem<Base>::sumcount(0);
template<class Base> size_t CountItem<Base>::mulcount(0);

}

#endif
