#ifndef UTIL_INPUT_GENERATOR_H
#define UTIL_INPUT_GENERATOR_H

#include <memory/mapping.h>
#include <memory/pointer.h>

namespace Matrix {
template<class Storage, class T> class Classic;
template<class Storage, class T> class Z;
}

namespace Util {
namespace Generator {

template<class Array, class Generator> void generate(Array& array, Generator generator, size_t count) {
	for (size_t i(0); i < count; ++ i)
		generator(array[i]);
}

template<class Storage, class T, class Generator> void generate(Matrix::Classic<Storage, T>& matrix, Generator generator) {
	matrix.iterate(generator);
}

template<class Storage, class T, class Generator> void generate(Matrix::Z<Storage, T>& matrix, Generator generator) {
	matrix.iterate(generator);
}

template<class T> void random(T& result) {
	result.random();
}

template<class Storage, class T> void random(Matrix::Classic<Storage, T>& matrix) {
	generate(matrix, random<T>);
}

template<class Storage, class T> void random(Matrix::Z<Storage, T>& matrix) {
	generate(matrix, random<T>);
}

template<> void random<int>(int& result);
template<> void random<unsigned>(unsigned& result);
template<> void random<char>(char& result);
template<> void random<double>(double& result);

}
}

#endif
