#ifndef UTIL_RINGBUF_H
#define UTIL_RINGBUF_H

#include <memory/mapping.h>
#include <memory/pointer.h>

namespace Util {

template<class Storage, class T> class RingBuf {
	private:
		RingBuf(const RingBuf<Storage, T> &other);
		RingBuf<Storage, T> &operator =(const RingBuf<Storage, T> &other);
	public:
		RingBuf(const Memory::Pointer<Storage, T> &buffer_, size_t size_, size_t filled = 0) :
			mapping(NULL),
			buffer(buffer_),
			head(buffer_),
			tail(buffer_ + filled),
			size(size_),
			bufferFull(false)
		{
			if (filled == size_) {
				bufferFull = true;
				tail = buffer;
			}
		}
		RingBuf(size_t size_) :
			mapping(new Memory::Mapping<Storage>(size_ * sizeof(T), "ring buffer")),
			buffer(*mapping),
			head(buffer),
			tail(buffer),
			size(size_),
			bufferFull(false)
		{ }
		~ RingBuf() {
			delete mapping;
		}
		bool empty() const {
			return (head == tail) && !bufferFull;
		}
		bool full() const {
			return bufferFull;
		}
		void put(const T& element) {
			*head = element;
			move(head);
			if (head == tail)
				bufferFull = true;
		}
		const T& get() {
			bufferFull = false;
			Memory::Pointer<Storage, T> orig(tail);
			move(tail);
			return *orig;
		}
		T& peek() {
			return *tail;
		}
	private:
		void move(Memory::Pointer<Storage, T> &p) {
			++ p;
			if (p == buffer + size)
				p = buffer;
		}
		Memory::Mapping<Storage> *mapping;
		Memory::Pointer<Storage, T> buffer, head, tail;
		size_t size;
		bool bufferFull;
};

};

#endif
