#ifndef UTIL_STACK_H
#define UTIL_STACK_H

#include <memory/pointer.h>
#include <memory/mapping.h>

namespace Util {

namespace Stacks {

template<class Storage, class T, class Base> class Mixin : public Base {
	private:
		typedef Memory::Pointer<Storage, unsigned char> RawPtr;
		typedef Memory::Pointer<Storage, T> Ptr;
		Ptr baseptr() {
			RawPtr bp(Base::buffer);
			bp += Base::base;
			return Ptr::casted(bp);
		}
	public:
		// Delegate parameters
		Mixin() : Base() { }
		template<class Param> Mixin(Param &p) : Base(p) { };
		void push(const T &item) {
			// FIXME: Used only with items smaller than blocksize
			if (Base::base + (Base::used + 1) * sizeof(T) > Base::buffer.size()) {
				Base::buffer.resize(Base::buffer.size() + Storage::blocksize());
			}
			baseptr()[Base::used ++] = item;
		}
		T &pop() {
			return baseptr()[-- Base::used];
		}
		T &top() {
			return baseptr()[Base::used - 1];
		}
		bool empty() const {
			return !Base::used;
		}
};

template<class Storage, class T> class Data {
	public:
		Data() :
			buffer(Storage::blocksize(), "stack"),
			used(0)
		{ }
	public:
		// Should be protected, but declaring template friends kind of doesn't work
		Memory::Mapping<Storage> buffer;
		size_t used;
		static const size_t base = 0;
};

template<class Storage, class T> class SharedData {
	private:
		void round() {
			base += sizeof(T) - 1;
			base /= sizeof(T);
			base *= sizeof(T);
		}
	public:
		template<class Src> SharedData(Data<Storage, Src> &other) :
			buffer(other.buffer),
			used(0),
			base(other.base + other.used * sizeof(Src))
		{
			round();
		}
		template<class Src> SharedData(SharedData<Storage, Src> &other) :
			buffer(other.buffer),
			used(0),
			base(other.base + other.used * sizeof(Src))
		{
			round();
		}
		Memory::Mapping<Storage> &buffer;
		size_t used;
		size_t base;
};

}

template<class Storage, class T> class Stack : public Stacks::Mixin<Storage, T, Stacks::Data<Storage, T> > { };

template<class Storage, class T> class SharedStack : public Stacks::Mixin<Storage, T, Stacks::SharedData<Storage, T> > {
	public:
		template<class Param> SharedStack(Param &p) : Stacks::Mixin<Storage, T, Stacks::SharedData<Storage, T> >(p) { }
};

}

#endif
