#!/usr/bin/perl
use common::sense;
use Data::Dumper;

my %levels;
my $currentLevel;

sub finish() {
	my @keys = keys %levels;
	die "Leftover levels (@keys)" if @keys;
}

sub init($$) {
	my ($n, $m) = @_;
}

sub checkCorrectLevel($$$$) {
	my ($level, $inputSize, $outputSize, $depth) = @_;
}

my $childKind;

sub activate($$) {
	my ($level, $phase) = @_;
	$currentLevel = $levels{$level} // die "Level '$level' in phase $phase not found";
	# TODO: Check the phase is allowed/expected now
	$currentLevel->{phase} = $phase;
	if ($phase eq 'C') {
		$childKind = 'contract';
	}
}

sub newLevel($$$$$) {
	my ($level, $phase, $inputSize, $outputSize, $depth) = @_;
	$levels{$level} = {
		phase => $phase,
		I_size => $inputSize,
		O_size => $outputSize,
		depth => $depth,
		id => $level,
	};
	push @{$currentLevel->{children}}, $levels{$level} if $currentLevel;
}

my $dumpLevel;

sub startDump($$) {
	my ($level, $part) = @_;
	die "Two dumps at once ($dumpLevel->{id} and $level)" if $dumpLevel;
	$dumpLevel = $levels{$level} // die "Missing level '$level' to dump";
	#$dumpLevel->{$part}->{older} = $dumpLevel->{$part}->{current};
	delete $dumpLevel->{$part}->{current};
}

sub endDump($) {
	my ($part) = @_;
	# TODO: Check the level
	if ($part eq 'O') {
		my %right = map {
			$_->[1] => 1
		} @{$dumpLevel->{O}->{current}};
		for (@{$dumpLevel->{O}->{current}}) {
			if ($right{$_->[0]}) {
				warn Dumper $currentLevel, $dumpLevel;
				die "$_->[0] on both sides in $dumpLevel->{id}";
			}
		}
	}
	if (($childKind eq 'split') && (@{$currentLevel->{children}} == 3)) { #One of them is the 'Free' level
		# Check there are no duplicities
		my %first = map {
			$_->[0] => 1, $_->[1] => 1
		} @{$currentLevel->{children}->[0]->{I}->{current}};
		die "Duplicate item in lists" if grep { $first{$_->[0]} || $first{$_->[1]} } @{$currentLevel->{children}->[2]->{I}->{current}};
		my %second = map {
			$_->[0] => 1, $_->[1] => 1
		} @{$currentLevel->{children}->[2]->{I}->{current}};
		$currentLevel->{sideparts} = [\%first, undef, \%second];
	} elsif (($currentLevel->{phase} eq 'C') && ($currentLevel->{id} == 47446)) {
		warn Dumper $currentLevel, $dumpLevel;
	} elsif ($currentLevel->{phase} eq 'C') {
		my %forbidden;
		my $parts = $currentLevel->{sideparts};
		for my $child (qw(0 2)) {
			for (@{$currentLevel->{children}->[$child]->{O}->{current}}) {
				for my $index (0, 1) {
					die "Invalid number $_->[$index] in $currentLevel->{id}" unless $parts->[$child]->{$_->[$index]};
				}
				$forbidden{$_->[0]} = 1;
			}
		}
		my $last = $currentLevel->{children}->[-1];
		for my $edge (@{$last->{I}->{current}}) {
			for (0, 1) {
				die "Forbidden side ($edge->[0], $edge->[1]) on $currentLevel->{id} with index $_" if $parts->[2 * $_]->{$edge->[$_]};
				if ($forbidden{$edge->[$_]}) {
					warn Dumper $currentLevel;
					die "This is contracted away! ($edge->[$_])";
				}
			}
		}
	}
	undef $dumpLevel;
}

sub dumpData($$$$) {
	my ($part, $endpoint1, $endpoint2, $hints) = @_;
	die "No dump active" unless $dumpLevel;
	push @{$dumpLevel->{$part}->{current}}, [$endpoint1, $endpoint2, $hints];
}

sub done($) {
	my ($level) = @_;
	die "Done a different level ($level instead of $currentLevel->{id})" if $level != $currentLevel->{id};
	delete $currentLevel->{children};
	undef $currentLevel;
	delete $levels{$level};
}

sub doSplit() {
	$childKind = 'split';
}

sub doThrough() {
	undef $childKind;
}

sub doRecursion() {
	$childKind = 'recursion';
}

sub doNoSplit() {
	undef $childKind;
}

open my $process, '-|', './bin/components' or die "Could not run the program ($!)";

while (<$process>) {
	chomp;
	s/^DATA: // or next;
	if (/^Start (\d+) (\d+)$/) {
		init $1, $2;
	} elsif (/^Level (\d+) phase (.) sizes (\d+) (\d+) depth (-?\d+) activate$/) {
		checkCorrectLevel $1, $3, $4, $5;
		activate $1, $2;
	} elsif (/^Level (\d+) phase (.) sizes (\d+) (\d+) depth (-?\d+) $/) {
		newLevel $1, $2, $3, $4, $5;
	} elsif (/^dump (\d+) (.)$/) {
		startDump $1, $2;
	} elsif (/^([IO]) dump end$/) {
		endDump $1;
	} elsif (/^([IO]) (\d+) (\d+) (-?\d+)$/) {
		dumpData $1, $2, $3, $4;
	} elsif (/^done (\d+)/) {
		done $1;
	} elsif (/^Split$/) {
		doSplit;
	} elsif (/^Through$/) {
		doThrough;
	} elsif (/^Recursion$/) {
		doRecursion;
	} elsif (/^No-split$/) {
		doNoSplit;
	} elsif (/^Finish$/) {
		finish;
	} else {
		warn "Unknown line '$_'";
	}
}

finish;
