#!/usr/bin/perl
use common::sense;
use Getopt::Long qw(:config auto_abbrev permute);

my $template = '';
my $prog = './tools/data.pl';
my $src = 'output';
my $title = 'Algorithms';
my $output = 'graph.png';
my $algorithms = 'mergeFifo mergeStack qsort heap funnel funnelHybrid funnelFullHybrid';
my $xlabel = 'Number of elements (log)';
my $ylabel = 'Time (s, log)';
my $log;
my $open;
my $tex;

GetOptions(log => \$log, open => \$open, tex => \$tex, (map(("$_=s" => eval "\\\$$_"), qw(template prog src title output algorithms xlabel ylabel)))) or die "Could not parse options";

my @algs = split /\s+/, $algorithms;

system "$prog $template --alg=$_ $src >$_.data.$$" foreach @algs;

open my $plot, '|gnuplot' or die "Could not start gnuplot ($!)\n";
select $plot;
$\ = ";\n";
if ($tex) {
	print "set terminal pslatex color 10 size 16cm,9cm";
} else {
	print "set terminal png size 1024, 768";
	print "set output '$output'";
}
#print "set title '$title'";
print "set log xy" if $log;
#print 'set format xy "\\ifthenelse{%L<-3}{$10^{%L}}{\\ifthenelse{%L>3}{$10^{%L}$}{$%g$}}$"' if $log;
print 'set format xy "$10^{%L}$"' if $log;
print "set xlabel '$xlabel'";
print "set ylabel '$ylabel'";
print "set key left top";
my @colors = qw(red blue black orchid green brown purple olivegreen orange #83ffd5 #007f00 #8a0000);
my $cnum;
print "plot ", (join ', ', (map "'$_.data.$$' title '$_' with linespoints lt 1 lc rgb \"".$colors[$cnum ++ % scalar @colors]."\"", sort @algs));
close $plot;
unlink map "$_.data.$$", @algs;
exec ('geeqie', $output) if $open;
